#!/usr/bin/env python
# encoding: utf-8
from __future__ import division
import random
import pytest
import bpf4


N = 1000
EPSILON = 0.000001


def random_floats(n, rnge):
    x0, x1 = rnge
    return [random.random() * (x1 - x0) + x0 for n in range(n)]

    
def assert_almost_equal(a, b):
    assert abs(a - b) < EPSILON


def test_linear_arithm():
    a = bpf4.core.Linear((0, 10), (0, 100))
    cases = random_floats(n=N, rnge=(-10, 20))
    for x in cases:
        try:
            print(x)
            assert a(x) + 10 == (a + 10)(x)
            assert a(x) - 13.5 == (a - 13.5)(x)
            assert a(x) * 0.5 == (a * 0.5)(x)
            assert a(x) / 8.3 == (a / 8.3)(x)
            assert a(x) * a(x) == (a * a)(x)
            assert a(x) + a(x) == (a + a)(x)
            assert a(x) - a(x) * 2 == (a - a(x) * 2)(x) == (a - a * 2)(x)
            assert a(x) / (a(x) - 1) == (a / (a - 1))(x)
            assert a(x) ** x == (a ** x)(x)
            assert float(a(x) > 5) == float((a > 5)(x))
            assert float(a(x) >= 5) == float((a >= 5)(x))
            assert float(a(x) < 1.3) == float((a < 1.3)(x))
            assert float(a(x) <= 1.3) == float((a <= 1.3)(x))
            assert abs(a(x) * -1) == abs(a * -1)(x)
            assert_almost_equal(  a(x) * 10, (a + a + a + a + a + a + a + a + a + a)(x)  )    
        except ZeroDivisionError:
            pass


def test_linear_integration():
    def square(b):
        surface = (b.x1 - b.x0) * b(b.x0)
        assert b.integrate() == surface
        assert b.integrate_between(b.x0, b.x1) == surface
        assert b.integrate_between(b.x0, (b.x0 + b.x1)/2) == surface/2

    def bpfs():
        y = 1.5
        b = bpf4.linear(0, y, 1, y, 10, y, 10.01, y, 10.010001, y, 20, y)
        yield b

        yield bpf4.linear(-1, 10, 1, 10, 10000, 10)

    for b in bpfs():
        square(b)

# Run tests with 
# py.test /path/to/this/file



"""
## High-level API for bpf4

The API allows a high-level and flexible interface to the core of bpf4,
which is implemented in cython for efficiency.

### API vs core

These three curves, *a*, *b* and *c* define the same linear break-point-function
The first two definitions, *a* and *b*, use the high-level API, which allows for
points to be defined as a flat sequence, as tuples of (x, y). The *core* classes
need to be instantiated with two arrays of *x* and *y* values, as in *c*

```python

from bpf4 import *
a = linear(0, 0, 1, 2.5, 3, 10)
b = linear((0, 0), (1, 2.5), (3, 10))
c = core.Linear([0, 1, 3], [0, 2.5, 10])
```

"""
from __future__ import annotations

from . import core
from . import util


def linear(*args) -> core.Linear:
    """
    Construct a Linear bpf.

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    A bpf can be constructed in multiple ways, all of which result
    in the same Linear instance:

    ```python

    linear(x0, y0, x1, y1, ...)
    linear((x0, y0), (x1, y1), ...)
    linear({x0:y0, x1:y1, ...})
    ```

    Example
    -------

    ```python
    from bpf4 import *
    a = linear([0, 2, 3.5, 10], [0.1, 0.5, -3.5,  4])
    a.plot()
    ```
    ![](assets/Linear.png)
    """
    X, Y, kws = util.parseargs(*args)
    if kws:
        raise ValueError("linear does not take any keyword")
    return core.Linear(X, Y)


def expon(*args, **kws) -> core.Expon:
    """
    Construct an Expon bpf (a bpf with exponential interpolation)

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`
        exp: the exponent to use
        numiter: Number of iterations. A higher number accentuates the effect

    A bpf can be constructed in multiple ways:

    ```python
    expon(x0, y0, x1, y1, ..., exp=exponent)
    expon(exponent, x0, y0, x1, y1, ...)
    expon((x0, y0), (x1, y1), ..., exp=exponent)
    expon({x0:y0, x1:y1, ...}, exp=exponent)
    ```

    Example
    -------

    ```python
    from bpf4 import *
    import matplotlib.pyplot as plt
    numplots = 5
    fig, axs = plt.subplots(2, numplots, tight_layout=True, figsize=(20, 8))
    for i in range(numplots):
        exp = i+1
        expon(0, 0, 1, 1, exp=exp).plot(show=False, axes=axs[0, i])
        expon(0, 0, 1, 1, exp=1/exp).plot(show=False, axes=axs[1, i])
        axs[0, i].set_title(f'{exp=}')
        axs[1, i].set_title(f'exp={1/exp:.2f}')

    plot.show()
    ```
    ![](assets/expon-grid.png)

    """
    X, Y, kws = util.parseargs(*args, **kws)
    assert "exp" in kws
    return core.Expon(X, Y, **kws)


def halfcos(*args, exp=1, numiter=1, **kws) -> core.Halfcos:
    """
    Construct a half-cosine bpf (a bpf with half-cosine interpolation)

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`
        exp: the exponent to use
        numiter: Number of iterations. A higher number accentuates the effect

    A bpf can be constructed in multiple ways:

    ```python

    halfcos(x0, y0, x1, y1, ...)
    halfcos((x0, y0), (x1, y1), ...)
    halfcos({x0:y0, x1:y1, ...})
    ```

    ```python
    a = halfcos([0, 1, 3, 10], [0.1, 0.5, 3.5,  1])
    b = halfcos(*a.points(), exp=2)
    c = halfcos(*a.points(), exp=0.5)
    fig, axes = plt.subplots(1, 3, figsize=(16, 4), tight_layout=True)
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1], show=False)
    c.plot(axes=axes[2])
    ```
    ![](assets/Halfcos.png)


    """
    X, Y, kws = util.parseargs(*args, **kws)
    return core.Halfcos(X, Y, exp=exp, numiter=numiter, **kws)


halfcosexp = halfcos


def halfcosm(*args, **kws) -> core.Halfcosm:
    """
    Halfcos interpolation with symmetric exponent

    When used with an exponent, the exponent is inverted for downwards
    segments `(y1 > y0)`

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`
        exp: exponent to apply prior to cosine interpolation. The higher the exponent, the
            more skewed to the right the shape will be
        numiter: Number of iterations. A higher number accentuates the effect

    Returns:
        (core.Halfcosm) A bpf with symmetric cosine interpolation


    A bpf can be constructed in multiple ways:

    ```python

    halfcosm(x0, y0, x1, y1, ..., exp=2.0)
    halfcosm(2.0, x0, y0, x1, y1, ...)    # The exponent can be placed first
    halfcosm((x0, y0), (x1, y1), ...)
    halfcosm({x0:y0, x1:y1, ...})
    ```

    ```python
    from bpf4 import *
    a = halfcosm(0, 0.1,
                 1, 0.5,
                 3, 3.5,
                 10, 1, exp=2)
    b = halfcosm(*a.points(), exp=2)
    fig, axes = plt.subplots(1, 2, figsize=(16, 4))
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1])
    ```
    ![](assets/Halfcosm.png)


    """
    X, Y, kws = util.parseargs(*args, **kws)
    return core.Halfcosm(X, Y, **kws)


def spline(*args) -> core.Spline:
    """
    Construct a cubic-spline bpf

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    Returns:
        (core.Spline) A Spline bpf


    A bpf can be constructed in multiple ways:

    ```python
    spline(x0, y0, x1, y1, ...)
    spline((x0, y0), (x1, y1), ...)
    spline({x0:y0, x1:y1, ...})
    ```

    ```python
    from bpf4 import *
    a = smooth(0, 0.1, 1, 0.5, 3, -3.5, 10, 1)
    b = spline(*a.points())
    fig, axes = plt.subplots(1, 2, figsize=(12, 4))
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1])
    ```
    ![](assets/Spline.png)
    """
    X, Y, kws = util.parseargs(*args)
    return core.Spline(X, Y)


def uspline(*args) -> core.USpline:
    """
    Construct a univariate cubic-spline bpf

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    Returns:
        (core.USpline) A USpline bpf

    A bpf can be constructed in multiple ways:

    ```python
    uspline(x0, y0, x1, y1, ...)
    uspline((x0, y0), (x1, y1), ...)
    uspline({x0:y0, x1:y1, ...})
    ```

    ```python
    from bpf4 import *
    a = spline(0, 0.1, 1, 0.5, 3, -3.5, 10, 1)
    b = uspline(*a.points())

    fig, axes = plt.subplots(1, 2, figsize=(12, 4), sharey=True, tight_layout=True)
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1])
    ```
    ![](assets/Uspline.png)

    !!! info "See Also"

        * [spline](#spline)
        * [pchip](#pchip)

    """
    X, Y, kws = util.parseargs(*args)
    return core.USpline(X, Y)


def nointerpol(*args) -> core.NoInterpol:
    """
    A bpf with floor interpolation

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    A bpf can be constructed in multiple ways:

        nointerpol(x0, y0, x1, y1, ...)
        nointerpol((x0, y0), (x1, y1), ...)
        nointerpol({x0:y0, x1:y1, ...})

    ```python
    a = linear([0, 1, 3, 10], [0.1, 0.5, 3.5,  1])
    b = nointerpol(*a.points())
    c = nearest(*a.points())
    fig, axes = plt.subplots(1, 3, figsize=(15, 4), tight_layout=True)
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1], show=False)
    c.plot(axes=axes[2])
    ```
    ![](assets/NoInterpol.png)
    """
    X, Y, kws = util.parseargs(*args)
    return core.NoInterpol(X, Y, **kws)


def nearest(*args) -> core.Nearest:
    """
    A bpf with floor interpolation

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
                a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
                or two arrays `xs` and `ys`

    A bpf can be constructed in multiple ways:

        nearest(x0, y0, x1, y1, ...)
        nearest((x0, y0), (x1, y1), ...)
        nearest({x0:y0, x1:y1, ...})

    ```python
    a = linear([0, 1, 3, 10], [0.1, 0.5, 3.5,  1])
    b = nointerpol(*a.points())
    c = nearest(*a.points())
    fig, axes = plt.subplots(1, 3, figsize=(15, 4), tight_layout=True)
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1], show=False)
    c.plot(axes=axes[2])
    ```
    ![](assets/NoInterpol.png)
    """
    X, Y, kws = util.parseargs(*args)
    return core.Nearest(X, Y, **kws)


def smooth(*args, numiter=1) -> core.Smooth:
    """
    A bpf with smoothstep interpolation.

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`
        numiter: determines the number of smoothstep steps applied
            (see https://en.wikipedia.org/wiki/Smoothstep)

    Returns:
        (core.Smooth) A bpf with smoothstep interpolation

    ```python

    from bpf4.api import *
    a = smooth((0, 0.1), (1, 0.5), (3, -3.5), (10, 1))
    a.plot()
    ```
    ![](assets/Smooth.png)

    !!! info "See Also"

        * [smoother](#smoother)
    """
    X, Y, kws = util.parseargs(*args)
    return core.Smooth(X, Y, numiter=numiter)


def smoother(*args) -> core.Smoother:
    """
    A bpf with smootherstep interpolation

    This bpf uses Perlin's variation on smoothstep,
    see https://en.wikipedia.org/wiki/Smoothstep)

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    Returns:
        (core.Smoother) A bpf with smootherstep interpolation


    ```python
    from bpf4 import *
    a = smooth(0, 0.1,
               1, 0.5,
               3, -3.5,
               10, 1)
    b = smoother(*a.points())
    fig, axes = plt.subplots(1, 2, figsize=(12, 4))
    a.plot(axes=axes[0], show=False)
    b.plot(axes=axes[1])
    ```
    ![](assets/Smoother.png)
    """
    X, Y, kws = util.parseargs(*args)
    return core.Smoother(X, Y)


def multi(*args):
    """
    A bpf with a per-pair interpolation

    Example
    -------

    ```python

    # (0,0) --linear-- (1,10) --expon(3)-- (2,3) --expon(3)-- (10, -1) --halfcos-- (20,0)

    multi(0, 0,   'linear'
          1, 10,  'expon(3)',
          2, 3,   # assumes previous interpolation
          10, -1, 'halfcos'
          20, 0)

    # also the following syntax is possible
    multi((0, 0, 'linear')
          (1, 10, 'expon(3)'),
          (2, 3),
          (10, -1, 'halfcos'),
          (20, 0))
    ```
    """
    xs, ys, interpolations = util.multi_parseargs(args)
    return core.Multi(xs, ys, interpolations)


def pchip(*args):
    """
    Monotonic Cubic Hermite Intepolation

    Args:
        args: either a flat list of coordinates in the form `x0, y0, x1, y1, ...`,
            a list of tuples `(x0, y0), (x1, y1), ...`, a dict `{x0:y0, x1:y1, ...}`
            or two arrays `xs` and `ys`

    A bpf can be constructed in multiple ways:

    ```python

    pchip(x0, y0, x1, y1, ...)
    pchip((x0, y0), (x1, y1), ...)
    pchip({x0:y0, x1:y1, ...})


    >>> a = core.Smoother([0, 1, 3, 10, 12, 12.5], [0.1, 0.5, -3.5,  1, 4.5, -1])
    >>> b = core.Spline(*a.points())
    >>> c = pchip(*a.points())

    >>> fig, axes = plt.subplots(1, 3, figsize=(16, 4), sharey=True, tight_layout=True)
    >>> a.plot(axes=axes[0], show=False)
    >>> b.plot(axes=axes[1], show=False)
    >>> c.plot()
    ```
    ![](assets/pchip.png)
    """
    from . import pyinterp
    xs, ys, kws = util.parseargs(*args)
    return pyinterp.Pchip(xs, ys, **kws)


def stack(*bpfs) -> core.Stack:
    """
    A bpf representing a stack of bpf

    Within a Stack, a bpf does not have outbound values. When evaluated
    outside its bounds the bpf below is used, iteratively until the
    lowest bpf is reached. Only the lowest bpf is evaluated outside its
    bounds

    Args:
        bpfs: a sequence of bpfs

    Returns:
        (core.Stack) A stacked bpf


    Example
    -------

    ```python
    # Interval    bpf
    # [0, 3]      a
    # (3, 4]      b
    # (4, 10]     c

    from bpf4 import *
    import matplotlib.pyplot as plt
    a = linear(0, 0, 3, 1)
    b = linear(2, 9, 4, 10)
    c = halfcos(0, 0, 10, 10)
    s = core.Stack((a, b, c))

    ax = plt.subplot(111)
    a.plot(color="#f00", alpha=0.4, axes=ax, linewidth=4, show=False)
    b.plot(color="#00f", alpha=0.4, axes=ax, linewidth=4, show=False)
    c.plot(color="#f0f", alpha=0.4, axes=ax, linewidth=4, show=False)
    s.plot(axes=ax, linewidth=2, color="#000", linestyle='dotted')
    ```
    ![](assets/stack2.png)

    """
    if len(bpfs) == 1 and isinstance(bpfs[0], (list, tuple)):
        bpfs = bpfs[0]
    return core.Stack(bpfs)


def blendshape(shape0: str, shape1: str, mix, points) -> core.BpfInterface:
    """
    Create a bpf blending two interpolation forms

    Args:
        shape0: a description of the first interpolation
        shape1: a description of the second interpolation
        mix (float | core.BpfInterface): blend factor.
            A value between 0 (use only `shape0`)
            and 1 (use only `shape1`). A value of `0.5` will result in
            an average between the first and second interpolation kind.
            Can be a bpf itself, returning the mix value at any x value
        points: either a tuple `(x0, y0, x1, y1, ...)` or a tuple `(xs, ys)`
            where *xs* and *ys* are lists/arrays containing the *x* and *y*
            coordinates of the points

    Returns:
        (core.BpfInterface) A bpf blending two different interpolation kinds

    Example
    -------

    ```python

    from bpf4 import *
    a = blendshape('halfcos(2.0)', 'linear', mix=0.5, points=(0, 0, 1, 1))
    halfcos(0, 0, 1, 1, exp=2).plot(color='red')
    linear(0, 0, 1, 1).plot(color='blue')
    a.plot(color='green')
    ```
    ![](assets/blend1.png)
    """
    X, Y, kws = util.parseargs(*points)
    a = util.makebpf(shape0, X, Y)
    b = util.makebpf(shape1, X, Y)
    return core.blend(a, b, mix)



import numpy as np
from typing import Union, Iterator, Callable
import numpy.typing as npt


class BpfPointsError(ValueError): ...

class BpfInversionError(ValueError): ...


class BpfInterface:
    def ntodx(self, N: int) -> float: ...
    def dxton(self, dx: float) -> int: ...
    def bounds(self) -> tuple[float, float]: ...

    @property
    def x0(self) -> float: ...


    @property
    def x1(self) -> float: ...

    def __add__(self, b: BpfInterface | float) -> BpfInterface: ...

    def __sub__(self, b: BpfInterface | float) -> BpfInterface: ...

    def __mul__(self, b: BpfInterface | float) -> BpfInterface: ...

    def __div__(self, b: BpfInterface | float) -> BpfInterface: ...

    def __truediv__(self, b: BpfInterface | float) -> BpfInterface: ...

    def __pow__(self, b: BpfInterface | float, module: float) -> BpfInterface: ...

    def __neg__(self: BpfInterface) -> BpfInterface: ...

    def __mod__(self, other: BpfInterface | float) -> BpfInterface: ...

    def __abs__(self) -> BpfInterface: ...

    def __or__(self, b: BpfInterface|float) -> BpfInterface: ...

    def __rshift__(self, b: BpfInterface|float) -> BpfInterface: ...

    def __lshift__(self, b: BpfInterface|float) -> BpfInterface: ...

    def __xor__(self, b: BpfInterface|float) -> BpfInterface: ...

    def _get_points_for_rendering(self, n: int = -1) -> tuple[np.ndarray, np.ndarray]: ...

    def render(self, xs: int | np.ndarray | list[float], interpolation: str = 'linear') -> np.ndarray: ...

    def sampled(self, dx: float, interpolation='linear') -> BpfInterface: ...

    def plot(self, kind: str = 'line', n: int = -1, show: bool = True, axes=None, **keys) -> None: ...

    def sample_between(self, x0: float, x1: float, dx: float, out: np.ndarray | float | None = None) -> np.ndarray: ...

    def sampled_between(self, x0: float, x1: float, dx: float, interpolation='linear') -> BpfInterface: ...

    def mapn_between(self, n: int, x0: float, x1: float, out: np.ndarray | None = None) -> np.ndarray: ...

    def map(self, xs: int | np.ndarray, out: np.ndarray | None = None) -> np.ndarray: ...

    def concat(self, other: BpfInterface)-> BpfInterface: ...

    def round(self) -> BpfInterface: ...

    def rand(self) -> BpfInterface: ...

    def cos(self) -> BpfInterface: ...

    def sin(self) -> BpfInterface: ...

    def ceil(self) -> BpfInterface: ...

    def expon(self) -> BpfInterface: ...

    def floor(self) -> BpfInterface: ...

    def tanh(self) -> BpfInterface: ...

    def abs(self) -> BpfInterface: ...

    def sqrt(self) -> BpfInterface: ...

    def acos(self) -> BpfInterface: ...

    def asin(self) -> BpfInterface: ...

    def tan(self) -> BpfInterface: ...

    def sinh(self) -> BpfInterface: ...

    def log10(self) -> BpfInterface: ...

    def log(self, base: float = 2.718281828459045) -> BpfInterface: ...

    def m2f(self) -> BpfInterface: ...

    def f2m(self) -> BpfInterface: ...
    def db2amp(self) -> BpfInterface: ...
    def amp2db(self) -> BpfInterface: ...
    def clip(self, y0: float = float('-inf'), y1: float = float('inf')) -> BpfInterface: ...
    def derivative(self) -> BpfInterface: ...
    def integrated(self) -> BpfInterface: ...
    def integrate(self) -> float: ...
    def integrate_between(self, x0: float, x1: float, N: int = 0) -> float: ...
    def mean(self) -> float: ...
    def zeros(self, h: float = 0.01, N: int = 0,
              x0: float = float('nan'), x1: float = float('nan'), maxzeros: int = 0) -> list[float]: ...
    def max(self, b: BpfInterface|float) -> BpfInterface: ...
    def min(self, b: BpfInterface|float) -> BpfInterface: ...
    def __call__(self, other: float) -> float: ...

    def keep_slope(self, epsilon: float = 0.) -> BpfInterface: ...

    def outbound(self, y0: float, y1: float) -> BpfInterface: ...

    def apply(self, func) -> BpfInterface: ...

    def preapply(self, func) -> BpfInterface: ...

    def periodic(self) -> BpfInterface: ...

    def stretched(self, rx: float, fixpoint: float = 0.) -> BpfInterface: ...

    def fit_between(self, x0: float, x1: float) -> BpfInterface: ...

    def shifted(self, dx: float) -> BpfInterface: ...

    def inverted(self) -> BpfInterface: ...

    @classmethod
    def fromseq(cls, *points: float|tuple[float, float], **kws) -> BpfInterface: ...

    def copy(self) -> BpfInterface: ...


class BpfBase(BpfInterface):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...

    @property
    def descriptor(self) -> str: ...

    def mapn_between(self, n: int, xstart: float, xend: float, out: np.ndarray|None = None) -> np.ndarray: ...

    def points(self) -> tuple[np.ndarray, np.ndarray]: ...

    def clone_with_new_data(self, xs: np.ndarray, ys: np.ndarray) -> BpfInterface: ...

    def insertpoint(self, x: float, y: float) -> BpfInterface: ...

    def removepoint(self, x: float) -> BpfInterface: ...

    def segments(self) -> Iterator[tuple[float, float, float]]: ...

    @property
    def exp(self) -> float: ...


class Linear(BpfBase):

    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


    def inverted(self) -> Linear: ...

    def flatpairs(self) -> np.ndarray: ...


class Smooth(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class Smoother(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class Halfcos(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]],
                 exp: float = 1.0, numiter: int =1): ...


class Halfcosm(Halfcos):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]],
                 exp: float = 1.0, numiter: int =1): ...


class Expon(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]],
                 exp: float = 1.0, numiter: int =1): ...


class Exponm(Expon):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]],
                 exp: float = 1.0, numiter: int =1): ...


class NoInterpol(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class Nearest(BpfBase):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class Sampled(BpfInterface):
    def __init__(self,
                 samples: np.ndarray,
                 dx: float,
                 x0: float = 0.,
                 interpolation: str = 'linear'): ...

    @property
    def ys(self) -> np.ndarray: ...


    @property
    def samplerate(self) -> float: ...

    @property
    def xs(self) -> np.ndarray: ...


    @property
    def dx(self) -> float: ...

    def set_interpolation(self, interpolation: str) -> Sampled: ...


    def flatpairs(self) -> np.ndarray: ...



class Spline(BpfInterface):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class USpline(BpfInterface):
    def __init__(self,
                 xs: Union[np.ndarray, list[float]],
                 ys: Union[np.ndarray, list[float]]
                 ): ...


class Slope(BpfInterface):

    @property
    def slope(self) -> float: ...

    @slope.setter
    def slope(self, value: float) -> None: ...

    @property
    def offset(self) -> float: ...

    @offset.setter
    def offset(self, value: float) -> None: ...



def _FunctionWrap(f: Callable[[float], float], bounds=(INFNEG, INF)) -> BpfInterface: ...


class _BpfCompose(BpfInterface): ...


class _BpfConcat(BpfInterface): ...


class _BpfBlend(BpfInterface): ...


class Multi(BpfInterface):
    def __init__(self, xs: npt.ArrayLike, ys: npt.ArrayLike, interpolations: list[str]) -> None: ...


class Const(BpfInterface):
    def __init__(self, value: float) -> None: ...


class _MultipleBpfs(BpfInterface):

    def __init__(self, bpfs: list[BpfInterface]) -> None: ...


class Max(_MultipleBpfs): ...

class Min(_MultipleBpfs): ...

class Stack(_MultipleBpfs): ...


def brentq(bpf: BpfInterface, x0: float, xa: float, xb: float, xtol=9.9999999999999998e-13,
           rtol=4.4408920985006262e-16, max_iter=100) -> tuple[float,int]: ...


def blend(a: BpfInterface, b: BpfInterface, mix=0.5) -> BpfInterface: ...



def bpf_zero_crossings(b: BpfInterface,
                       h: float = 0.01, N: int = 0,
                       x0: float = float('nan'),
                       x1: float = float('nan'),
                       maxzeros: int = 0
                       ) -> list[float]: ...



from . import core
from . import util


from .core import (
    BpfInterface,
    BpfBase,
    Linear,
    Smooth,
    Smoother,
    Halfcos,
    Expon,
    Nearest,
    Sampled,
    Slope,
    Spline,
    Max,
    Min,
    blend,
    Const
)

from .api import (
    linear,
    expon,
    halfcos,
    spline,
    halfcosm,
    uspline,
    nointerpol,
    nearest,
    smooth,
    smoother,
    multi,
    pchip,
)


from .util import asbpf


__all__ = [
    'core',
    'util',
    'BpfInterface',
    'BpfBase',
    'Linear',
    'Smooth',
    'Smoother',
    'Halfcos',
    'Expon',
    'Nearest',
    'Sampled',
    'Slope',
    'Spline',
    'Max',
    'Min',
    'blend',
    'Const',
    'linear',
    'expon',
    'halfcos',
    'spline',
    'halfcosm',
    'uspline',
    'nointerpol',
    'nearest',
    'smooth',
    'smoother',
    'multi',
    'pchip',
    'asbpf',
]



"""
Utilities for bpf4
"""
from __future__ import annotations
import operator as _operator
import os as _os
import itertools as _itertools
from functools import reduce
from math import isnan, sqrt
import numpy as np

from . import core

from typing import Sequence


_CSV_COLUMN_NAMES = ('x', 'y', 'interpolation', 'exponent')


_CONSTRUCTORS = {
    'linear': core.Linear,
    'expon': core.Expon,
    'halfcos': core.Halfcos,
    'nointerpol': core.NoInterpol,
    'spline': core.Spline,
    'uspline': core.USpline,
    'slope': core.Slope,
    'nearest': core.Nearest,
    'halfcosm': core.Halfcosm,
    'smooth': core.Smooth,
    'smoother': core.Smoother
}


def _isiterable(obj) -> bool:
    try:
        iter(obj)
        return not isinstance(obj, str)
    except TypeError:
        return False


def bpf_to_dict(bpf: core.BpfInterface) -> dict:
    """
    Convert a bpf to a dict with the following format

    Args:
        bpf: the bpf to convert to a dict

    Returns:
        The bpf as a dictionary

    Example
    ~~~~~~~

    .. code-block:: python

        >>> b = bpf.expon(3.0, 0, 0, 1, 10, 2, 20)
        >>> bpf_to_dict(b)
        {
            'interpolation': 'expon(3.0)',
            'points': [0, 0, 1, 10, 20, 20]  # [x0, y0, x1, y1, ...]
        }

        >>> b = bpf.multi(0, 0, 'linear',
                        1, 10, 'expon(2)',
                        3, 25)
        >>> bpf_to_dict(b)
        {
            'interpolation': 'multi',
            'segments': [
                [0, 0, 'linear'],
                [1, 10, 'expon(2)',
                [3, 25, '']]
        }
    """
    try:
        segments = list(bpf.segments())
    except AttributeError:
        raise TypeError("this kind of BPF cannot be translated. It must be rendered first.")
    d = {}
    interpolation = segments[0][2]
    # x y interpolation exp

    def normalize_segment(segment):
        """
        a segment as returned by .segments() is [x, y, interpl, exp]
        we want [x, y, interpol(exp)]
        """
        if len(segment) == 4:
            x, y, interpol, exp = segment
            if not interpol and exp == 0:
                # the last segment of a multi bpf
                segment = [x, y]
            else:
                interpol = "{interpol}({exp})".format(interpol=interpol, exp=exp)
                segment = [x, y, interpol]
        return segment

    if not all(segment[2] == interpolation for segment in segments[:-1]):
        # multi
        d['interpolation'] = 'multi'
        segments = [normalize_segment(seg) for seg in segments]
        d['segments'] = segments
    else:
        try:
            exp = bpf.exp
            if exp != 1:
                assert "(" not in interpolation
                interpolation = "{interp}({exp})".format(interp=interpolation, exp=exp)
        except AttributeError:
            pass
        d['interpolation'] = interpolation
        points = []
        for segment in segments:
            points.append(segment[0])
            points.append(segment[1])
        d['points'] = points
    return d


def dict_to_bpf(d: dict) -> core.BpfInterface:
    """
    Convert a dict to a bpf

    Args:
        d: the dictionary to convert to a bpf

    Returns:
        the converted bpf


    **Format 1**

    ```python

    bpf = {
        'interpolation': 'expon(2)',
        10: 0.1,
        15: 1,
        25: -1
    }
    ```

    **Format 2**

    ```python
    bpf = {
        'interpolation': 'linear',
        'points': [x0, y0, x1, y1, ...]
    }
    ```

    **Format 2b**

    ```python
    bpf = {
        'interpolation': 'linear',
        'points': [(x0, y0), (x1, y1), ...]
    }
    ```

    **Format 3 (multi)**

    ```python

    bpf = {
        'interpolation': 'multi',
        'segments': [
            [x0, y0, 'descr0'],
            [x1, y1, 'descr1'],
            ...
            [xn, yn, '']
        ]
    }
    ```
    """

    # check format
    if 'points' in d:
        # format 2
        interpolation = d.get('interpolation', 'linear')
        points = d['points']
        if isinstance(points[0], (int, float)):
            # format 2a
            X = points[::2]
            Y = points[1::2]
        elif isinstance(points[0], (list, tuple)):
            X = [point[0] for point in points]
            Y = [point[1] for point in points]
        else:
            raise ValueError(f"Invalid points format, expected list or tuple, got {points}")
        return makebpf(interpolation, X, Y)
    elif 'segments' in d and d['interpolation'] == 'multi':
        segments = d['segments']
        X = [s[0] for s in segments]
        Y = [s[1] for s in segments]
        interpolations = [s[2] for s in segments[:-1]]
        return core.Multi(X, Y, interpolations)
    else:
        # format 1
        interpolation = d.get('interpolation', 'linear')
        points = [(k, v) for k, v in d.items() if isinstance(k, (int, float))]
        points.sort()
        X, Y = list(zip(*points))
        return makebpf(interpolation, X, Y)


def asbpf(obj, bounds=(-np.inf, np.inf)) -> core.BpfInterface:
    """
    Convert obj to a bpf

    Args:
        obj: the object to convert to a bpf (a function, a dict, a constant
            or a bpf, in which case it is returned as is)
        bounds: the lower and upper bounds where this bpf is defined

    Returns:
        the resulting bpf
    """
    if isinstance(obj, core.BpfInterface):
        return obj
    elif callable(obj):
        return core._FunctionWrap(obj, bounds)
    elif hasattr(obj, '__float__'):
        return core.Const(float(obj))
    else:
        raise TypeError("can't wrap %s" % str(obj))


def parseargs(*args, **kws) -> tuple[list[float], list[float], dict]:
    """
    Convert the args and kws to the canonical form (xs, ys, kws)

    Returns:
        (tuple[list[float], list[float], dict]) A tuple `(xs, ys, kws)`

    Raises ValueError if failed

    All the following variants result in the same result:

    ```python

    x0, y0, x1, y1, …, exp=0.5
    (x0, y0), (x1, y1), …, exp=0.5
    {x0:y0, x1:y1, …}, exp=0.5
    [x0, x1, …], [y0, y1, …], exp=0.5

    Result: [x0, x1, …], [y0, y1, …], {exp:0.5}
    ```
    """
    L = len(args)
    if L == 0:
        raise ValueError("no arguments given")
    elif L == 1:
        d = args[0]
        assert isinstance(d, dict)
        items = list(d.items())
        items.sort()
        xs, ys = list(zip(*items))
    elif L == 2:
        if not all(map(_isiterable, args)):
            raise ValueError(f"parsing error: 2 args, expected a seq. "
                             f"[(x0, y1), (x1, y1)] or (xs, ys) but got {args}")
        if len(args[0]) > 2:   # <--  (xs, ys)
            xs, ys = args
        else:                  # <--  ((x0, y0), (x1, y1), ...)
            xs, ys = list(zip(*args))
    elif not any(map(_isiterable, args)):
        # (x0, y0, x1, y1, ...)
        if L % 2 == 0:  # even
            xs = args[::2]
            ys = args[1::2]
        else:
            if kws:
                raise ValueError(f"Uneven number of args. No keywords allowed in "
                                 f"this case, but got {kws}")
            kws = {'exp': args[0]}
            xs = args[1::2]
            ys = args[2::2]
    elif all(map(_isiterable, args)):   # <-- ((x0, y0), (x1, y1), ...)
        xs, ys = list(zip(*args))
    else:
        raise ValueError("could not parse arguments")
    return xs, ys, kws



def parsedescr(descr: str, validate=True) -> tuple[str, dict]:
    """
    Parse interpolation description


    | descr                   | output                             |
    |-------------------------|------------------------------------|
    | linear                  | linear, {}                         |
    | expon(0.4)              | expon, {'exp': 0.4}                |
    | halfcos(2.5, numiter=1) | halfcos, {'exp':2.5, 'numiter': 1} |

    """
    if "(" not in descr:
        classname = descr
        kws = {}
    else:
        classname, rest = descr.split("(")
        assert rest[-1] == ")"
        rest = rest[:-1]
        parts = rest.split(",")
        kws = {}
        for part in parts:
            if "=" in part:
                key, value = part.split("=")
                kws[key] = float(value)
            else:
                assert 'exp' not in kws
                kws['exp'] = float(part)
    if validate:
        assert classname in _CONSTRUCTORS
    return classname, kws


def makebpf(descr: str, X: Sequence[float], Y: Sequence[float]) -> core.BpfInterface:
    """
    Create a bpf from the given descriptor and points

    Args:
        descr: a string descriptor of the interpolation ("linear", "expon(xx)", ...)
        X: the array of xs
        Y: the array of ys

    Returns:
        the created bpf
    """
    interpolkind, kws = parsedescr(descr)
    bpfclass = _CONSTRUCTORS.get(interpolkind)
    if bpfclass is None:
        raise ValueError(f"descr {descr} is not valid")
    return bpfclass(X, Y, **kws)


def multi_parseargs(args) -> tuple[list[float], list[float], list[str]]:
    """
    Parse args of a multi bpf

    Given a list of args of the form (x0, y0, interpol) or (x0, y0) (or a flat
    version thereof), fills the possibly missing interpolation descriptions
    and returns a tuple `(xs, ys, interpolations)`

    Returns:
        a tuple (xs, ys, interpolations), where len(interpolations) == len(xs) - 1
    """
    xs = []
    ys = []
    interpolations = []
    last_interpolation = 'linear'
    if all(isinstance(arg, (int, float, str)) for arg in args):
        # a flat list
        accum = 0
        for arg in args:
            if accum == 0:
                xs.append(arg)
                accum += 1
            elif accum == 1:
                ys.append(arg)
                accum += 1
            else:
                if isinstance(arg, str):
                    interpolations.append(arg)
                    last_interpolation = arg
                    accum = 0
                else:
                    interpolations.append(last_interpolation)
                    xs.append(arg)
                    accum = 1
    else:
        # it is of the type (x0, y0, interpolation), (x1, y1), ...
        assert all(isinstance(arg, tuple) and 2 <= len(arg) <= 3 for arg in args)

        for arg in args[:-1]:
            if len(arg) == 2:
                x, y = arg
                interp = last_interpolation
            elif len(arg) == 3:
                x, y, interp = arg
            xs.append(x)
            ys.append(y)
            interpolations.append(interp)
        x, y = args[-1]
        xs.append(x)
        ys.append(y)
    assert all(isinstance(x, (int, float)) for x in xs)
    assert all(isinstance(y, (int, float)) for x in ys)
    assert all(isinstance(i, str) for i in interpolations)
    assert len(xs) == len(ys) == len(interpolations)+1
    return xs, ys, interpolations


def max_(*elements) -> core.Max:
    """
    Return a bpf representing the max over the given elements

    Args:
        elements: each element can be a bpf or a scalar

    Returns:
        a Max bpf
    """
    bpfs = list(map(asbpf, elements))
    return core.Max(*bpfs)


def min_(*elements) -> core.Min:
    """
    Return a bpf representing the min over elements

    Args:
        elements: each can be a bpf or a scalar

    Returns:
        a Min bpf
    """
    bpfs = list(map(asbpf, elements))
    return core.Min(*bpfs)


def sum_(*elements):
    """
    Return a bpf representing the sum of elements

    Args:
        elements: each can be a bpf or a scalar

    Returns:
        a bpf representing the sum of all elements

    """
    bpfs = list(map(asbpf, elements))
    return reduce(_operator.add, bpfs)


def select(which: core.BpfInterface, bpfs: Sequence[core.BpfInterface], shape='linear') -> core._BpfSelect:
    """
    Create a new bpf which interpolates between adjacent bpfs

    Args:
        which: returns at any x, which bpf from bpfs should return the result
        bpfs: a list of bpfs
        shape: interpolation shape between consecutive bpfs

    Returns:
        a BpfSelect

    **Example**

    ```python

    >>> which = nointerpol(0, 0, 5, 1)
    >>> bpfs = [linear(0, 0, 10, 10), linear(0, 10, 10, 0)]
    >>> s = select(which, bpfs)
    >>> s(1)     # at time=1, the first bpf will be selected
    0
    ```
    """
    return core._BpfSelect(asbpf(which), list(map(asbpf, bpfs)), shape)


def dumpbpf(bpf: core.BpfInterface, fmt='yaml') -> str:
    """
    Dump the data of this bpf as human readable text


    Args:
        bpf: the bpf to dump
        fmt: the format, one of 'csv', 'yaml', 'json'

    Returns:
        the text representation according to the format

    The bpf can then be reconstructed via `loadbpf`

    """
    if fmt == 'csv':
        if outfile is None:
            raise ValueError("need an outfile to dump to CSV")
        return bpf_to_csv(bpf, outfile)
    elif fmt == 'json':
        return bpf_to_json(bpf, outfile)
    elif fmt == 'yaml':
        return bpf_to_yaml(bpf, outfile)
    else:
        raise ValueError(f"Format {fmt} not supported")


def concat_bpfs(bpfs: list[core.BpfInterface]) -> core._BpfConcat:
    """
    Concatenate these bpfs together, one after the other
    """
    bpfs2 = [bpfs[0]]
    x0, x1 = bpfs[0].bounds()
    xs = [x0]
    for bpf in bpfs[1:]:
        bpf2 = bpf.fit_between(x1, x1 + (bpf._x1 - bpf._x0))
        bpfs2.append(bpf2)
        xs.append(bpf2._x0)
        x0, x1 = bpf2.bounds()
    return core._BpfConcat(xs, bpfs2)


def warped(bpf: core.BpfInterface, dx:float=None, numpoints=1000) -> core.Sampled:
    """
    Represents the curvature of a linear space.

    The result is a warped bpf so that:

    ```
    position_bpf | warped_bpf = corresponding position after warping
    ```

    Args:
        dx: the accuracy of the measurement
        numpoints: if dx is not given, the bpf is sampled `numpoints` times
            across its bounds

    Returns:
        (core.Sampled) The warped bpf


    Example
    -------

    Find the theoretical position of a given point according to a
    probability distribution

    ```python
    >>> from bpf4 import *
    >>> import matplotlib.pyplot as plt
    >>> distribution = halfcos(0,0, 0.5,1, 1, 0)
    >>> w = util.warped(distribution)
    >>> distribution.plot()
    >>> w.plot()

    ```
    ![](assets/warped.png)

    Now plot the histrogram of the warped bpf. It should resemble the
    original distribution
    ```python
    plt.hist(w.map(10000), bins=200, density=True)
    ```
    ![](assets/warped-hist.png)

    Using another distribution, notice that the histogram follows the
    distribution again:

    ```python
    distribution = halfcos(0,0, 0.8,1, 1, 0, exp=3.)
    w = util.warped(distribution)
    distribution.plot()
    w.plot()
    _ = plt.hist(w.map(10000), bins=200, density=True)[2]
    ```
    ![](assets/warped-hist2.png)
    """
    from scipy.optimize import brentq as _brentq

    x0, x1 = bpf.bounds()
    if dx is None:
        dx = (x1 - x0) / numpoints
    integrated = bpf.integrated()[::dx]
    integrated_at_x1 = integrated(bpf.x1)
    # N = int((x1 + dx - x0) / dx + 0.5)
    xs = np.arange(x0, x1+dx, dx)
    ys = np.ones_like(xs) * np.nan
    for i in range(len(xs)):
        try:
            ys[i] = _brentq(integrated - xs[i]*integrated_at_x1, x0, x1)
        except:
            pass
    return core.Sampled(ys, dx=dx, x0=x0)


def _minimize(bpf, N: int, func=min, debug=False) -> float | None:
    x0, x1 = bpf.bounds()
    xs = np.linspace(x0, x1, N)
    from scipy.optimize import brent
    mins = [brent(bpf, brack=(xs[i], xs[-i])) for i in range(int(len(xs) * 0.5 + 0.5))]
    mins2 = [(bpf(m), m) for m in mins]  # if x0 <= m <= x1]
    if debug:
        print(mins2)
    if mins2:
        return float(func(mins2)[1])
    return None


def minimum(bpf: core.BpfInterface, N=10) -> float | None:
    """
    Find the x where bpf(x) is minimized

    Args:
        bpf: the bpf to analyze
        N: the number of estimates

    Returns:
        the *x* value where bpf(x) is minimized. Returns `None` if
        no minimum found

    """
    return _minimize(bpf, N, min)


def maximum(bpf: core.BpfInterface, N=10) -> float | None:
    """
    return the x where bpf(x) is the maximum of bpf

    Args:
        bpf: the bpf to analyze

    Returns:
        the *x* value where bpf(x) is the maximum. Returns `None` if
        no maximum found

    """
    return _minimize(-bpf, N, min)


def rms(bpf: core.BpfInterface, rmstime=0.1) -> core.BpfInterface:
    """
    The rms of this bpf

    Args:
        bpf: the bpf
        rmstime: the time to calculate the rms over

    Returns:
        a bpf representing the rms of this bpf at any x coord
    """
    bpf2 = bpf**2

    def func(x):
        return sqrt(bpf2.integrate_between(x, x+rmstime) / rmstime)
    return asbpf(func, bounds=(bpf.x0, bpf.x1))


def rmsbpf(samples: np.ndarray, sr: int, dt=0.01, overlap=2, smoothen=0.) -> core.BpfInterface:
    """
    Return a bpf representing the rms of the given samples as a function of time

    Args:
        samples: the audio samples
        sr: the sample rate
        dt: analysis time period
        overlap: overlap of analysis frames
        smoothen: if given, the returned bpf is smoothen using the given value as window

    Returns:
        a sampled bpf if not smoothening operation is performed, or a linear
        bpf if smoothening is required
    """
    s = samples
    period = int(sr * dt + 0.5)
    hopsamps = period // overlap
    dt2 = hopsamps / sr
    numperiods = len(s) // hopsamps
    data = np.empty((numperiods,), dtype=float)
    for i in range(numperiods):
        idx0 = i * hopsamps
        chunk = s[idx0:idx0+period]
        data[i] = sqrt(np.mean(np.square(chunk)))
    out = bpf4.core.Sampled(data, x0=0, dx=dt2)
    if smoothen:
        out = smoothen(out, window=smoothen)


def calculate_projection(x0, x1, p0, p1):
    """
    Calculate a projection needed to map the interval x0:x1 to p0:p1

    Returns:
        (tuple[float, float, float]) A tuple (rx, dx, offset)
    """
    rx = (x1-x0) / (p1-p0)
    dx = x0
    offset = p0
    return rx, dx, offset


def projection_fixedpoint(rx, dx, offset):
    """
    Returns the fixed point given the projection parameters

    x2 = (x-offset)*rx + dx

    For a fixed point, x2 == x
    """
    return (-offset*rx + dx)/(1-rx)



def binarymask(mask: str | list[int],
               durs: Sequence[float]=None,
               offset=0.,
               cycledurs=True
               ) -> core.NoInterpol:
    """
    Creates a binary mask

    Args:
        mask: a mask string ('x'=1, '-'=0) or a sequence of states (a state is either 0 or 1)
        durs: a sequence of durations (default=[1])

    Returns:
        (core.NoInterpol) A NoInterpol bpf representing the binary mask

    **Example**

    ```python

    >>> mask = binarymask("x--x-x---")
    ```

    """
    if durs is None:
        durs = [1]
    if cycledurs:
        durs = _itertools.cycle(durs)
    else:
        assert len(durs) == len(mask)

    def binvalue(x):
        d = {'x':1, 'o':0, '-':0, 1:1, 0:0, '1':1, '0':0}
        return d.get(x)

    mask = [binvalue(x) for x in mask]
    mask.append(mask[-1])
    assert all(x is not None for x in mask)
    t = offset
    times = []
    for i, dur in zip(range(len(mask)), durs):
        times.append(t)
        t += dur
    return core.NoInterpol(times, mask)


def _pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = _itertools.tee(iterable)
    try:
        next(b)
    except StopIteration:
        pass
    return zip(a, b)


def jagged_band(xs: list[float], upperbpf: core.BpfInterface, lowerbpf=0, curve='linear'
                ) -> core.BpfInterface:
    """
    Create a jagged bpf between lowerbpf and upperbpf at the x values given

    At each x in xs the, the value is equal to lowerbpf, sweeping
    with curvature 'curve' to upperbpf just before the next x

    Example
    -------

    ```python

    from bpf4 import *
    import numpy as np
    a = expon(0, 0, 1, 10, exp=2)
    b = expon(0, 0, 1, 5, exp=4)
    j = util.jagged_band(list(np.arange(0, 1, 0.1)), a, b, curve='expon(1.5)')
    j.plot()
    ```
    ![](assets/jagged.png)
    """
    upperbpf = asbpf(upperbpf)
    lowerbpf = asbpf(lowerbpf)
    if not isinstance(xs, list):
        xs = list(xs)
    EPSILON = 1e-12
    fragments = []
    if xs[0] > upperbpf.x0 > float('-inf'):
        xs = [upperbpf.x0] + xs
    if xs[-1] < upperbpf.x1 < float('inf'):
        xs.append(upperbpf.x1)
    for x0, x1 in _pairwise(xs[:-1]):
        x1 = x1 - EPSILON
        fragment = makebpf(curve, [x0, x1], [lowerbpf(x0), upperbpf(x1)])[x0:x1].outbound(0, 0)
        fragments.append(fragment)
    x0 = xs[-2]
    x1 = xs[-1]
    fragments.append(makebpf(curve, [x0, x1], [lowerbpf(x0), upperbpf(x1)])[x0:x1].outbound(0, 0))
    return max_(*fragments)


def randombw(bw: float | core.BpfInterface,
             center: float | core.BpfInterface
             ) -> core.BpfInterface:
    """
    Create a random bpf

    Args:
        bw: a (time-varying) bandwidth
        center: the center of the random distribution

    Returns:
        a bpf

    if randombw is 0.1 and center is 1, the bpf will render values
    between 0.95 and 1.05

    !!! note

        This bpf will always be different, since the random numbers
        are calculated as needed. Sample it to freeze it to a known state.

    **Example**

    ```python

    >>> l = bpf.linear(0, 0, 1, 1)
    >>> r = bpf.util.randombw(0.1)
    >>> l2 = (l*r)[::0.01]
    ```
    """
    bw = asbpf(bw)
    return (bw.rand() + (center - bw*0.5))[bw.x0:bw.x1]


def blendwithfloor(b: core.BpfInterface, mix=0.5) -> core._BpfBlend:
    """
    Returns a blend of b with its minimum y value
    """
    return core.blend(b, asbpf(b(minimum(b))), mix)[b.x0:b.x1]


def blendwithceil(b, mix=0.5) -> core._BpfBlend:
    """
    Returns a blend of b with its maximum y value
    """
    return core.blend(b, asbpf(b(maximum(b))), mix)[b.x0:b.x1]


def smoothen(b: core.BpfInterface, window: float, N=1000, interpol='linear') -> core.BpfInterface:
    """
    Return a bpf representing a smooth version of b

    Args:
        b: a bpf
        window: the width (in x coords) of the smoothing window
        N: number of points to resample the bpf
        interpol: the interpolation to use. One of 'linear' or 'smooth'

    Returns:
        a bpf representing a smoother version of b

    ## Example

    ```python
    >>> import bpf4 as bpf
    >>> b = bpf.linear(0, 0, 0.1, 1, 0.2, 10, 0.3, 1, 0.5, 3, 0.8, -2)
    >>> bsmooth = bpf.util.smoothen(b, window=0.05)
    >>> axes = b.plot(show=False)
    >>> bsmooth.plot(axes=axes)
    ```

    ![](assets/smoothen.png)

    """
    dx = min((b.x1 - b.x0) / N, window/7)
    nwin = int(window / dx)
    box = np.ones(nwin)/nwin
    Y0 = b[::dx].ys

    Ypad = np.ones(shape=(nwin,), dtype='float') * Y0[-1]
    Y = np.concatenate((Y0, Ypad))
    Ysmooth = np.convolve(Y, box, mode="same")[:len(Y0)]
    X = np.linspace(b.x0, b.x1, len(Ysmooth))

    if interpol == 'linear':
        return core.Linear(X, Ysmooth)
    elif interpol == 'smooth':
        return core.Smooth(X, Ysmooth)
    else:
        raise ValueError(f"Interpolation '{interpol}' not supported here. "
                          "Possible values: linear, smooth")


def zigzag(b0: core.BpfInterface,
           b1: core.BpfInterface,
           xs: Sequence[float],
           shape='linear'
           ) -> core.BpfInterface:
    """
    Creates a curve formed of lines from b0(x) to b1(x) for each x in xs

    Args:
        b0: a bpf
        b1: a bpf
        xs: a seq. of x values to evaluate b0 and b1
        shape: the shape of each segment

    Returns:
        The resulting bpf

    ```

       *.
        *...  b0
         *  ...
         *     ...
          *       ....
           *          ...
            *         :  ...
             *        :*    ...
             *        : *      ...
              *       :  **       ...
               *      :    *         :*.
                *     :     *        : **...
                 *    :      *       :   *  ...
                 *    :       *      :    *    ...
                  *   :        *     :     **     .:.
                   *  :         *    :       *     :**..
                    * :          **  :        **   :  ****.
                     *:            * :          *  :      ****
        -----------  *:             *:           * :          ****
          b1       ---*--------------*---         **:             ****
                                         -----------*----------      .**
                                                               -----------
        x0            x1              x2                       x3

    ```
    """
    curves = []
    for x0, x1 in pairwise(xs):
        X = [x0, x1]
        Y = [b0(x0), b1(x1)]
        curve = bpf.util.makebpf(shape, X, Y)
        curves.append(curve)
    jointcurve = bpf.max_(*[c.outbound(0, 0) for c in curves])
    return jointcurve


def bpfavg(b: core.BpfInterface,
           dx: float
           ) -> core.BpfInterface:
    """
    Return a Bpf which is the average of b over the range `dx`

    Args:
        b: the bpf
        dx: the period to average *b* over

    Returns:
        a bpf representing the average of *b* along the bounds of
        *b* over a sliding period of *dx*
    """
    dx2 = dx/2
    avg = ((b<<dx2)+b+(b>>dx2))/3.0
    return avg[b.x0:b.x1]


def histbpf(b: core.BpfInterface, numbins=20, numsamples=400
            ) -> core.Linear:
    """
    Create a historgram of *b*

    Args:
        b: the bpf
        numbins: the number of bins
        numsamples: how many samples to take to determine the histogram
        interpolation: the kind of interpolation of the returned bpf

    Returns:
        a bpf mapping values to percentiles. The returned bpf can be inverted
        (see example) to map percentiles to values

    ## Example

    ```python

    >>> from sndfileio import *
    >>> import bpf4
    >>> samples, sr = sndread("path/to/soundfile.wav")
    >>> dbcurve = bpf4.util.rmsbpf(samples, sr=sr).amp2db()
    >>> dbval2hist = bpf4.util.histbpf(dbcurve)
    # Find the db percentile at a given time, this gives a measurement of the
    # relative strength of the sound at a given moment
    >>> dur = len(samples)/sr
    >>> percentile = dbval2hist(dur*0.5)
    0.312Z
    >>> dbhist2val = dbval2hist.inverted()

    ```

    This indicates that at the middle of the sound the amplitude is at percentile ~30

    """
    samples = b.map(numsamples)
    edges, hist = np.histogram(samples, bins=numbins)
    percentile = np.linspace(0, 1, len(hist))
    return core.Linear(hist, percentile)


def split_fragments(b: core.BpfBase, sep=float('nan')) -> list[core.BpfBase]:
    """
    Split a bpf into its fragments

    Fragments are defined by the separator, which is NAN by default. This separator
    splits the points in this bpf into fragments. A fragment must at least have two items

    Args:
        b: the bpf to split. This bpf must be a Linear, Sampled or any other BpfBase
            subclass (HalfCos, Smooth, etc.). For any other bpf the bpf needs to be
            sampled (`bpf[::period]`)
        sep: the separator to use

    Returns:
        a list of bpfs representing the fragments

    ## Example

    ```python

    >>> a = bpf.linear(0, 0, 1, 10, 2, 5, 3, 30, 4, nan, 5, nan, 6, 0, 7, 1, 8, 0.5, 9, nan, 10, 2, 11, 3)
    >>> split_fragments(a)
    [Linear[0.0:3.0], Linear[6.0:8.0], Linear[10.0:11.0]]
    ```
    """
    if not isinstance(b, core.BpfBase):
        raise TypeError(f"This function only works for subclasses of BpfBase (Linear, Sampled, etc.), "
                        f"got {b}.")
    xs, ys = b.points()
    parts = []
    lastpart: list[tuple[float, float]] | None = None
    if isnan(sep):
        for x, y in zip(xs, ys):
            if not isnan(y):
                if lastpart is None:
                    lastpart = []
                    parts.append(lastpart)
                lastpart.append((x, y))
            else:
                lastpart = None
    else:
        for x, y in zip(xs, ys):
            if y == sep:
                if lastpart is None:
                    lastpart = []
                    parts.append(lastpart)
                lastpart.append((x, y))
            else:
                lastpart = None
    cls = b.__class__
    bpfs = []

    for part in parts:
        xs, ys = zip(*part)
        if len(xs) <= 1:
            continue
        if cls is core.Sampled:
            bpfs.append(core.Sampled(ys, x0=xs[0], dx=xs[1]-xs[0]))
        else:
            bpfs.append(cls(xs, ys))
    return bpfs


def simplify_linear_coords(xs: np.ndarray, ys: np.ndarray, threshold=0., ratio=0.) -> tuple[np.ndarray, np.ndarray]:
    """
    Simplify the linear bpf defined by the points xs and ys

    Args:
        xs: the x coord array
        ys: the y coord array
        threshold: the simplification threshold. Points which deviate less than this
            value will be simplified. The first and last points are never simplified

    Returns:
        the simplified line as a tuple (xs, ys)
    """
    if len(xs) <= 2:
        return (xs, ys)
    import visvalingamwyatt as vw
    points = np.column_stack((xs, ys))
    simp = vw.Simplifier(points)
    if threshold:
        simplified_points = simp.by_threshold(threshold)
    elif ratio:
        simplified_points = simp.by_ratio(ratio)
    else:
        raise ValueError("Either threshold or ratio should be given")
    return simplified_points[:, 0], simplified_points[:, 1]



"""
BPF

A package implementing piece-wise interpolation functions in Cython
"""
from setuptools import setup, Extension
import numpy


setup(
    name = "bpf4",
    python_requires='>=3.9',
    ext_modules = [
        Extension(
            "bpf4.core",
            sources=["bpf4/core.pyx"],
            include_dirs = ["bpf4", numpy.get_include()]
        )
    ],
    install_requires = ['numpy>=1.8', 'matplotlib', 'scipy', 'visvalingamwyatt'],
    packages = ['bpf4'],
    # package_data={'bpf4': ['core.pyi', '__init__.pyi', 'py.typed']},
    package_data={'bpf4': ['__init__.pyi', 'py.typed']},
    include_package_data=True,
)
